import math
import pandas as pd
from tqdm import tqdm
import torch
from torch.utils.data import DataLoader, Subset
from pytorch_lightning import Trainer
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split

from .encoding import get_encoder
from .models import TextRegressionModel
from .utils import chunk_text, pad_chunks, TextRegressionDataset, collate_fn

class TextRegressor:
    """
    A text regression estimator following an sklearn-like API.
    
    This estimator takes in a pandas DataFrame containing a 'text' column and a 'y'
    column (with optional exogenous feature columns) and processes the text using configurable
    encoding and chunking, then applies a deep learning model (with RNN-based layers) to predict
    the target variable.
    """
    def __init__(self, 
                 encoder_model="sentence-transformers/all-MiniLM-L6-v2",
                 rnn_type="LSTM",
                 rnn_layers=2,
                 hidden_size=512,
                 bidirectional=True,
                 inference_layer_units=100,
                 chunk_info=None,
                 padding_value=0,
                 exogenous_features=None,
                 learning_rate=1e-3,
                 loss_function="mae",
                 max_steps=500,
                 early_stop_enabled=False,
                 patience_steps=None,
                 val_check_steps=50,
                 **kwargs):
        """
        Initialize the TextRegressor.
        
        Args:
            encoder_model (str): Pretrained encoder model identifier.
            rnn_type (str): Type of RNN to use ("LSTM" or "GRU").
            rnn_layers (int): Number of RNN layers.
            hidden_size (int): Hidden size for the RNN.
            bidirectional (bool): Whether to use bidirectional RNN.
            inference_layer_units (int): Units in the final linear inference layer.
            chunk_info (tuple, optional): (chunk_size, overlap) for splitting long texts.
            padding_value (int, optional): Padding value for text chunks.
            exogenous_features (list, optional): List of additional exogenous feature column names.
            learning_rate (float): Learning rate for the optimizer.
            loss_function (str): Loss function to use. Options: "mae", "smape", "mse", "rmse", "wmape", "mape".
            max_steps (int): Maximum number of training steps. Default is 500.
            early_stop_enabled (bool): Whether to enable early stopping. Default is False.
            patience_steps (int, optional): Number of steps with no improvement after which training will be stopped.
                If early_stop_enabled=True and not provided, defaults to 10.
            val_check_steps (int): Check validation every this many training steps. Default is 50.
            **kwargs: Additional keyword arguments.
        """
        self.encoder_model = encoder_model
        self.rnn_type = rnn_type
        self.rnn_layers = rnn_layers
        self.hidden_size = hidden_size
        self.bidirectional = bidirectional
        self.inference_layer_units = inference_layer_units
        self.chunk_info = chunk_info
        self.padding_value = padding_value
        self.exogenous_features = exogenous_features
        self.learning_rate = learning_rate
        self.loss_function = loss_function
        self.max_steps = max_steps
        self.early_stop_enabled = early_stop_enabled
        self.val_check_steps = val_check_steps

        if self.early_stop_enabled:
            self.patience_steps = patience_steps if patience_steps is not None else 10
        else:
            self.patience_steps = None
        
        # Initialize the encoder.
        self.encoder = get_encoder(self.encoder_model)
        
        # Placeholder for the deep learning model; will be initialized in fit.
        self.model = None
        
        # Placeholder for exogenous feature scaler if needed.
        self.exo_scaler = None
        
    def fit(self, df, batch_size=64, val_size=None, **kwargs):
        """
        Fit the TextRegressor model on the provided DataFrame.
        
        The DataFrame must have a 'text' column and a 'y' column.
        Optionally, it can include additional exogenous feature columns.
        
        Args:
            df (pandas.DataFrame): DataFrame containing 'text' and 'y' columns.
            batch_size (int): Batch size for training. Default is 64.
            val_size (float, optional): Percentage (between 0 and 1) of data to use for validation.
                Required if early_stop_enabled is True.
            **kwargs: Additional arguments for model training.
            
        Returns:
            self: Fitted estimator.
        """
        if not isinstance(df, pd.DataFrame):
            raise ValueError("Input must be a pandas DataFrame")
        if 'text' not in df.columns or 'y' not in df.columns:
            raise ValueError("DataFrame must have 'text' and 'y' columns")
        
        texts = df['text'].tolist()
        targets = df['y'].tolist()
        encoded_sequences = []
        
        # Process each text: chunk and encode.
        for text in tqdm(texts, desc="Processing texts"):
            chunks = chunk_text(text, self.chunk_info, encoder=self.encoder)
            chunks = pad_chunks(chunks, padding_value=self.padding_value)
            encoded_chunks = [self.encoder.encode(chunk) for chunk in chunks]
            # Ensure each encoded chunk is a torch.Tensor.
            encoded_chunks = [chunk if isinstance(chunk, torch.Tensor) else torch.tensor(chunk)
                              for chunk in encoded_chunks]
            encoded_sequences.append(encoded_chunks)
        
        # Process exogenous features if provided.
        if self.exogenous_features is not None:
            exo_data = df[self.exogenous_features].values
            self.exo_scaler = StandardScaler()
            exo_data_scaled = self.exo_scaler.fit_transform(exo_data)
            exo_list = [list(row) for row in exo_data_scaled]
        else:
            exo_list = None
        
        # Create the dataset.
        dataset = TextRegressionDataset(encoded_sequences, targets, exogenous=exo_list)
        
        # Split into training and validation sets if early stopping is enabled.
        if self.early_stop_enabled:
            if val_size is None:
                raise ValueError("When early_stop_enabled is True, you must specify val_size (a float between 0 and 1).")
            indices = list(range(len(dataset)))
            train_idx, val_idx = train_test_split(indices, test_size=val_size, random_state=42)
            train_subset = Subset(dataset, train_idx)
            val_subset = Subset(dataset, val_idx)
            train_loader = DataLoader(train_subset, batch_size=batch_size, collate_fn=collate_fn, shuffle=True)
            val_loader = DataLoader(val_subset, batch_size=batch_size, collate_fn=collate_fn, shuffle=False)
        else:
            train_loader = DataLoader(dataset, batch_size=batch_size, collate_fn=collate_fn, shuffle=True)
            val_loader = None
        
        # Compute the number of epochs based on max_steps and steps per epoch.
        steps_per_epoch = len(train_loader)
        computed_epochs = math.ceil(self.max_steps / steps_per_epoch)
        
        # Determine encoder output dimension.
        if hasattr(self.encoder, 'model') and hasattr(self.encoder.model, 'get_sentence_embedding_dimension'):
            encoder_output_dim = self.encoder.model.get_sentence_embedding_dimension()
        elif hasattr(self.encoder, 'output_dim'):
            encoder_output_dim = self.encoder.output_dim
        else:
            encoder_output_dim = 768
        
        # Initialize the PyTorch Lightning model.
        self.model = TextRegressionModel(
            rnn_type=self.rnn_type,
            rnn_layers=self.rnn_layers,
            hidden_size=self.hidden_size,
            bidirectional=self.bidirectional,
            inference_layer_units=self.inference_layer_units,
            exogenous_features=self.exogenous_features,
            learning_rate=self.learning_rate,
            loss_function=self.loss_function,
            encoder_output_dim=encoder_output_dim
        )
        
        # Configure callbacks.
        callbacks = []
        if self.early_stop_enabled:
            from pytorch_lightning.callbacks import EarlyStopping
            early_stop_callback = EarlyStopping(
                monitor="val_loss",
                patience=self.patience_steps,
                mode="min",
                verbose=True,
            )
            callbacks.append(early_stop_callback)
        
        # Adjust the validation check interval to avoid errors.
        if self.early_stop_enabled:
            val_check_interval = min(self.val_check_steps, len(train_loader))
        else:
            val_check_interval = None
        
        # Create Trainer with computed epochs, max_steps, and GPU auto-detection.
        trainer = Trainer(
            max_steps=self.max_steps,
            max_epochs=computed_epochs,
            accelerator="auto",
            devices="auto",
            val_check_interval=val_check_interval,
            callbacks=callbacks
        )
        
        if self.early_stop_enabled:
            trainer.fit(self.model, train_dataloaders=train_loader, val_dataloaders=val_loader)
        else:
            trainer.fit(self.model, train_dataloaders=train_loader)
        
        return self
    
    def predict(self, df, batch_size=64, **kwargs):
        """
        Predict continuous target values for new text data in the provided DataFrame.
        
        The DataFrame must have a 'text' column (and optionally exogenous feature columns).
        
        Args:
            df (pandas.DataFrame): DataFrame containing a 'text' column.
            batch_size (int): Batch size for prediction. Default is 64.
            **kwargs: Additional arguments for prediction.
            
        Returns:
            predictions (list): Predicted continuous values.
        """
        if not isinstance(df, pd.DataFrame):
            raise ValueError("Input must be a pandas DataFrame")
        if 'text' not in df.columns:
            raise ValueError("DataFrame must have a 'text' column")
        
        texts = df['text'].tolist()
        encoded_sequences = []
        
        # Process each text: chunk and encode.
        for text in tqdm(texts, desc="Processing texts"):
            chunks = chunk_text(text, self.chunk_info, encoder=self.encoder)
            chunks = pad_chunks(chunks, padding_value=self.padding_value)
            encoded_chunks = [self.encoder.encode(chunk) for chunk in chunks]
            encoded_chunks = [chunk if isinstance(chunk, torch.Tensor) else torch.tensor(chunk)
                              for chunk in encoded_chunks]
            encoded_sequences.append(encoded_chunks)
        
        # Process exogenous features if provided.
        if self.exogenous_features is not None:
            if self.exo_scaler is None:
                raise ValueError("Model was not fitted with exogenous features.")
            exo_data = df[self.exogenous_features].values
            exo_data_scaled = self.exo_scaler.transform(exo_data)
            exo_list = [list(row) for row in exo_data_scaled]
        else:
            exo_list = None
        
        # Create the dataset with dummy target values.
        dataset = TextRegressionDataset(encoded_sequences, [0] * len(encoded_sequences), exogenous=exo_list)
        
        # Create DataLoader for prediction.
        predict_loader = DataLoader(dataset, batch_size=batch_size, collate_fn=collate_fn, shuffle=False)
        
        from pytorch_lightning import Trainer
        trainer = Trainer(accelerator="auto", devices="auto")
        predictions = trainer.predict(self.model, dataloaders=predict_loader)
        
        # Flatten predictions (trainer.predict returns a list of batch outputs).
        flat_predictions = [pred.item() for batch in predictions for pred in batch]
        return flat_predictions
    
    def fit_predict(self, df, **kwargs):
        """
        Fit the model on the provided DataFrame and immediately predict on it.
        
        Args:
            df (pandas.DataFrame): DataFrame containing 'text' and 'y' columns.
            **kwargs: Additional arguments.
            
        Returns:
            predictions (list): Predicted continuous values.
        """
        self.fit(df, **kwargs)
        return self.predict(df, **kwargs)



import pandas as pd
import torch
import numpy as np

# --- Import functions/classes from the package ---
from textregress.encoding import (
    get_encoder,
    SentenceTransformerEncoder,
    TfidfEncoder,
    HuggingFaceEncoder
)
from textregress.utils import chunk_text, pad_chunks, collate_fn, TextRegressionDataset
from textregress.estimator import TextRegressor

# ---------------------------------
# Utility Function Tests (utils.py)
# ---------------------------------
def test_chunk_text():
    text = "one two three four five six seven eight nine"
    # Using chunk_info=(3, 1) should yield only full chunks.
    chunks = chunk_text(text, chunk_info=(3, 1), encoder=None)  # encoder not used in chunk_text
    expected = ["one two three", "three four five", "five six seven", "seven eight nine"]
    assert chunks == expected, f"Expected {expected}, got {chunks}"

def test_pad_chunks():
    chunks = ["a", "b", "c"]
    padded = pad_chunks(chunks, padding_value=0)
    assert padded == chunks, "pad_chunks should return input as-is"

def test_collate_fn():
    # Create a dummy batch of two samples. Each sample is a tuple: (encoded_sequence, target).
    sample1 = ([torch.ones(10), torch.ones(10) * 2], 1.0)
    sample2 = ([torch.ones(10) * 3, torch.ones(10) * 4], 2.0)
    batch = [sample1, sample2]
    
    padded_sequences, targets = collate_fn(batch)
    # Expected shape: (batch_size, max_seq_len, feature_dim) = (2, 2, 10)
    assert padded_sequences.shape == (2, 2, 10), f"Got shape {padded_sequences.shape}"
    assert targets.shape[0] == 2

def test_text_regression_dataset():
    # Create a dummy dataset with two samples, each with two encoded chunks.
    sample1 = [torch.zeros(5), torch.ones(5)]
    sample2 = [torch.ones(5) * 2, torch.ones(5) * 3]
    targets = [10.0, 20.0]
    dataset = TextRegressionDataset([sample1, sample2], targets)
    assert len(dataset) == 2, "Dataset should have two samples"
    sample, target = dataset[0]
    assert isinstance(sample, list), "Each sample should be a list"
    assert sample[0].shape[0] == 5, "Chunk tensor should have shape (5,)"
    assert target == 10.0

# ---------------------------------
# Encoder Tests (encoding.py)
# ---------------------------------
def test_tfidf_encoder():
    encoder = TfidfEncoder()
    texts = ["hello world", "test sentence"]
    encoder.fit(texts)
    vector = encoder.encode("hello world")
    # Expect a 1D numpy array.
    assert isinstance(vector, np.ndarray)
    assert vector.ndim == 1

def test_sentence_transformer_encoder():
    # This test loads the actual model; may take a few seconds.
    encoder = SentenceTransformerEncoder("sentence-transformers/all-MiniLM-L6-v2")
    vector = encoder.encode("Hello world")
    assert torch.is_tensor(vector)
    # Expected output dimension for all-MiniLM-L6-v2 is 384.
    assert vector.shape[-1] == 384, f"Expected 384, got {vector.shape[-1]}"

def test_huggingface_encoder():
    encoder = HuggingFaceEncoder("bert-base-uncased")
    vector = encoder.encode("Hello world")
    assert torch.is_tensor(vector)
    # bert-base-uncased typically produces embeddings of dimension 768.
    assert vector.shape[-1] == 768, f"Expected 768, got {vector.shape[-1]}"

# ---------------------------------
# Estimator Integration Tests (estimator.py)
# ---------------------------------
# Create a dummy encoder for fast testing.
class DummyEncoder:
    def encode(self, text):
        # Return a fixed tensor of ones with dimension 768.
        return torch.ones(768)

def test_estimator_fit_predict_without_exogenous():
    df = pd.DataFrame({
        "text": ["Test sentence one.", "Test sentence two."],
        "y": [1.0, 2.0]
    })
    # Monkey-patch get_encoder to return DummyEncoder.
    from textregress import encoding
    original_get_encoder = encoding.get_encoder
    encoding.get_encoder = lambda model_id: DummyEncoder()
    
    estimator = TextRegressor(
        encoder_model="sentence-transformers/all-MiniLM-L6-v2",
        max_steps=2,
        early_stop_enabled=False,
        loss_function="mae"
    )
    estimator.fit(df)
    preds = estimator.predict(df)
    assert isinstance(preds, list)
    assert len(preds) == len(df)
    for pred in preds:
        assert isinstance(pred, float)
    
    # Restore original function.
    encoding.get_encoder = original_get_encoder

def test_estimator_fit_predict_with_exogenous_and_early_stop():
    df = pd.DataFrame({
        "text": ["Exogenous test one.", "Exogenous test two.", "Exogenous test three."],
        "y": [1.0, 2.0, 3.0],
        "exo1": [0.1, 0.2, 0.3],
        "exo2": [10, 20, 30]
    })
    from textregress import encoding
    original_get_encoder = encoding.get_encoder
    encoding.get_encoder = lambda model_id: DummyEncoder()
    
    estimator = TextRegressor(
        encoder_model="sentence-transformers/all-MiniLM-L6-v2",
        exogenous_features=["exo1", "exo2"],
        max_steps=3,
        early_stop_enabled=True,
        val_check_steps=1,
        loss_function="mse"
    )
    estimator.fit(df, val_size=0.33)
    preds = estimator.predict(df)
    assert isinstance(preds, list)
    assert len(preds) == len(df)
    for pred in preds:
        assert isinstance(pred, float)
    
    encoding.get_encoder = original_get_encoder

# --------------------------------
# Run Tests if This Script is Main
# --------------------------------
if __name__ == "__main__":
    print("Running tests...")
    
    test_chunk_text()
    print("test_chunk_text passed.")
    
    test_pad_chunks()
    print("test_pad_chunks passed.")
    
    test_collate_fn()
    print("test_collate_fn passed.")
    
    test_text_regression_dataset()
    print("test_text_regression_dataset passed.")
    
    test_tfidf_encoder()
    print("test_tfidf_encoder passed.")
    
    test_sentence_transformer_encoder()
    print("test_sentence_transformer_encoder passed.")
    
    test_huggingface_encoder()
    print("test_huggingface_encoder passed.")
    
    test_estimator_fit_predict_without_exogenous()
    print("test_estimator_fit_predict_without_exogenous passed.")
    
    test_estimator_fit_predict_with_exogenous_and_early_stop()
    print("test_estimator_fit_predict_with_exogenous_and_early_stop passed.")
    
    print("All tests passed!")



from sentence_transformers import SentenceTransformer
from sklearn.feature_extraction.text import TfidfVectorizer
import torch
from transformers import AutoTokenizer, AutoModel

def get_encoder(model_identifier):
    """
    Factory function to obtain an encoder based on the model identifier.
    
    Args:
        model_identifier (str): Identifier for the encoding model.
        
    Returns:
        An encoder object with an `encode` method.
    """
    model_identifier_lower = model_identifier.lower()
    if "tfidf" in model_identifier_lower:
        return TfidfEncoder()
    elif "sentence-transformers" in model_identifier_lower:
        return SentenceTransformerEncoder(model_identifier)
    else:
        # Use a generic HuggingFace encoder for any other model identifier.
        return HuggingFaceEncoder(model_identifier)

class SentenceTransformerEncoder:
    """
    Encoder using the SentenceTransformer library.
    """
    def __init__(self, model_name):
        self.model = SentenceTransformer(model_name)
    
    def encode(self, text):
        """
        Encode text (or a text chunk) into a vector representation.
        
        Args:
            text (str): The text to encode.
            
        Returns:
            vector (torch.Tensor): Encoded vector.
        """
        return self.model.encode(text, convert_to_tensor=True)

class TfidfEncoder:
    """
    Encoder using TFIDF Vectorizer.
    
    Note:
        The TFIDF encoder must be fitted on the full text corpus before encoding chunks.
    """
    def __init__(self):
        self.vectorizer = TfidfVectorizer()
        self.fitted = False
    
    def fit(self, texts):
        """
        Fit the TFIDF vectorizer on a list of texts.
        
        Args:
            texts (list): List of text documents.
        """
        self.vectorizer.fit(texts)
        self.fitted = True
    
    def encode(self, text):
        """
        Encode text using the fitted TFIDF vectorizer.
        
        Args:
            text (str): The text to encode.
            
        Returns:
            vector (ndarray): Encoded vector representation.
        """
        if not self.fitted:
            raise ValueError("TFIDF encoder has not been fitted yet.")
        return self.vectorizer.transform([text]).toarray()[0]

class HuggingFaceEncoder:
    """
    Encoder using HuggingFace transformers.
    
    This encoder loads a pre-trained model and tokenizer from HuggingFace and uses
    mean pooling over the token embeddings (from the last hidden state) as the sentence embedding.
    """
    def __init__(self, model_name):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModel.from_pretrained(model_name)
        self.model.eval()
    
    def encode(self, text):
        """
        Encode text (or a text chunk) into a vector representation using a HuggingFace model.
        
        Args:
            text (str): The text to encode.
            
        Returns:
            vector (torch.Tensor): Encoded vector.
        """
        inputs = self.tokenizer(text, return_tensors="pt", truncation=True, padding=True)
        with torch.no_grad():
            outputs = self.model(**inputs)
        # Mean pooling: average the token embeddings along the sequence dimension.
        embeddings = outputs.last_hidden_state.mean(dim=1).squeeze(0)
        return embeddings



from setuptools import setup, find_packages

setup(
    name="textregress",
    version="1.0.0",
    description="A package for performing linear regression on text data using configurable encoders and deep learning.",
    author="Jinhang Jiang, Weiyao Peng, Karthik Srinivasan",
    author_email="jinhang@asu.edu",
    packages=find_packages(),
    install_requires=[
        "pandas",
        "numpy",
        "scikit-learn",
        "tqdm",
        "torch",
        "pytorch-lightning",
        "sentence-transformers",
        "transformers",
    ],
    python_requires=">=3.6",
)



import torch
import torch.nn as nn
import pytorch_lightning as pl
from .loss import get_loss_function

class TextRegressionModel(pl.LightningModule):
    """
    PyTorch Lightning model for text regression.
    """
    def __init__(self, 
                 rnn_type="LSTM", 
                 rnn_layers=2, 
                 hidden_size=512,
                 bidirectional=True, 
                 inference_layer_units=100, 
                 exogenous_features=None,
                 learning_rate=1e-3,
                 loss_function="mae",
                 encoder_output_dim=768):
        """
        Initialize the TextRegressionModel.
        
        Args:
            rnn_type (str): Type of RNN to use ("LSTM" or "GRU").
            rnn_layers (int): Number of RNN layers.
            hidden_size (int): Hidden size for the RNN.
            bidirectional (bool): Whether to use a bidirectional RNN.
            inference_layer_units (int): Number of units in the final inference layer.
            exogenous_features (list, optional): List of exogenous feature names.
            learning_rate (float): Learning rate for the optimizer.
            loss_function (str): Loss function to use. Options: "mae", "smape", "mse", "rmse", "wmape", "mape".
            encoder_output_dim (int): Dimensionality of the encoder's output.
        """
        super(TextRegressionModel, self).__init__()
        self.save_hyperparameters()
        
        # Choose RNN class based on rnn_type.
        rnn_cls = nn.LSTM if rnn_type.upper() == "LSTM" else nn.GRU
        self.rnn = rnn_cls(
            input_size=encoder_output_dim,  # Dynamically set based on encoder.
            hidden_size=hidden_size,
            num_layers=rnn_layers,
            bidirectional=bidirectional,
            batch_first=True
        )
        
        # Calculate RNN output dimension.
        rnn_output_dim = hidden_size * (2 if bidirectional else 1)
        
        # Define the inference block.
        self.inference = nn.Linear(rnn_output_dim, inference_layer_units)
        
        # If exogenous features are provided, adjust the regressor input dimension.
        if exogenous_features is not None:
            exo_dim = len(exogenous_features)
            self.regressor = nn.Linear(inference_layer_units + exo_dim, 1)
        else:
            self.regressor = nn.Linear(inference_layer_units, 1)
        
        # Set up the loss function.
        self.criterion = get_loss_function(loss_function)
        self.learning_rate = learning_rate
        
    def forward(self, x, exogenous=None):
        """
        Forward pass.
        
        Args:
            x (torch.Tensor): Encoded text data of shape (batch_size, seq_len, feature_dim).
            exogenous (torch.Tensor, optional): Exogenous features.
            
        Returns:
            torch.Tensor: Regression output.
        """
        out, _ = self.rnn(x)
        # Use the output from the last time step.
        out = out[:, -1, :]
        out = self.inference(out)
        
        if exogenous is not None:
            out = torch.cat([out, exogenous], dim=1)
        
        output = self.regressor(out)
        return output
    
    def training_step(self, batch, batch_idx):
        if self.hparams.exogenous_features is not None:
            x, exogenous, y = batch
            y_hat = self(x, exogenous)
        else:
            x, y = batch
            y_hat = self(x)
        loss = self.criterion(y_hat.squeeze(), y.float())
        self.log('train_loss', loss)
        return loss

    def validation_step(self, batch, batch_idx):
        if self.hparams.exogenous_features is not None:
            x, exogenous, y = batch
            y_hat = self(x, exogenous)
        else:
            x, y = batch
            y_hat = self(x)
        loss = self.criterion(y_hat.squeeze(), y.float())
        self.log('val_loss', loss, prog_bar=True)
        return loss

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        if self.hparams.exogenous_features is not None:
            x, exogenous, _ = batch
            y_hat = self(x, exogenous)
        else:
            x, _ = batch
            y_hat = self(x)
        return y_hat.squeeze()
    
    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



import torch
from torch.nn.utils.rnn import pad_sequence

def chunk_text(text, chunk_info, encoder):
    """
    Splits the text into chunks.
    
    If chunk_info is provided, the text is split into overlapping chunks.
    Only full chunks (i.e. with exactly chunk_size words) are returned.
    Otherwise, the entire text is returned as a single chunk.
    
    Args:
        text (str): The input text.
        chunk_info (tuple or None): Tuple (chunk_size, overlap) or None.
        encoder: The encoder object.
        
    Returns:
        List[str]: List of text chunks.
    """
    if chunk_info:
        chunk_size, overlap = chunk_info
        words = text.split()
        chunks = []
        i = 0
        while i + chunk_size <= len(words):
            chunk = " ".join(words[i:i+chunk_size])
            chunks.append(chunk)
            i += max(chunk_size - overlap, 1)
        return chunks
    else:
        return [text]


def pad_chunks(chunks, padding_value=0):
    """
    Placeholder for padding chunks if necessary.
    Currently, this returns the chunks as-is.
    
    Args:
        chunks (list of str): List of text chunks.
        padding_value (int): Padding value.
        
    Returns:
        List[str]: Padded chunks.
    """
    return chunks

class TextRegressionDataset(torch.utils.data.Dataset):
    """
    Dataset for text regression.
    
    Each sample is a sequence of encoded chunks and a target value.
    Optionally includes exogenous features.
    """
    def __init__(self, encoded_sequences, targets, exogenous=None):
        """
        Args:
            encoded_sequences (list of list of torch.Tensor): Each inner list contains encoded chunks for a text sample.
            targets (list or array-like): Target values.
            exogenous (list, optional): List of exogenous feature vectors.
        """
        self.encoded_sequences = encoded_sequences
        self.targets = list(targets)
        self.exogenous = exogenous

    def __len__(self):
        return len(self.encoded_sequences)

    def __getitem__(self, idx):
        if self.exogenous is not None:
            return self.encoded_sequences[idx], self.exogenous[idx], self.targets[idx]
        else:
            return self.encoded_sequences[idx], self.targets[idx]

def collate_fn(batch):
    """
    Custom collate function to pad sequences of encoded chunks.
    
    Args:
        batch (list): List of samples, each sample is either (encoded_sequence, target) or (encoded_sequence, exogenous, target).
    
    Returns:
        Tuple: Padded sequences, (exogenous features if available), and targets.
    """
    # Check if exogenous features are provided.
    if len(batch[0]) == 3:
        sequences, exogenous, targets = zip(*batch)
    else:
        sequences, targets = zip(*batch)
        exogenous = None
    
    # Each item in sequences is a list of torch.Tensor; pad them along the sequence dimension.
    sequence_tensors = [torch.stack(seq) for seq in sequences]  # shape: (num_chunks, feature_dim)
    padded_sequences = pad_sequence(sequence_tensors, batch_first=True)  # shape: (batch_size, max_chunks, feature_dim)
    
    targets = torch.tensor(targets, dtype=torch.float32)
    
    if exogenous is not None:
        exogenous = torch.tensor(exogenous, dtype=torch.float32)
        return padded_sequences, exogenous, targets
    else:
        return padded_sequences, targets



import torch

def mae_loss(pred, target):
    """Mean Absolute Error"""
    return torch.mean(torch.abs(pred - target))

def mse_loss(pred, target):
    """Mean Squared Error"""
    return torch.mean((pred - target) ** 2)

def rmse_loss(pred, target):
    """Root Mean Squared Error"""
    return torch.sqrt(mse_loss(pred, target) + 1e-8)

def smape_loss(pred, target):
    """Symmetric Mean Absolute Percentage Error"""
    denominator = (torch.abs(pred) + torch.abs(target)) / 2.0 + 1e-8
    return torch.mean(torch.abs(pred - target) / denominator)

def mape_loss(pred, target):
    """Mean Absolute Percentage Error"""
    return torch.mean(torch.abs((target - pred) / (target + 1e-8))) * 100.0

def wmape_loss(pred, target):
    """Weighted Mean Absolute Percentage Error"""
    return torch.sum(torch.abs(target - pred)) / (torch.sum(torch.abs(target)) + 1e-8)

def get_loss_function(loss_name):
    """
    Factory function to get the loss function.
    
    Args:
        loss_name (str): One of "mae", "mse", "rmse", "smape", "mape", "wmape".
        
    Returns:
        A callable loss function.
    """
    loss_name = loss_name.lower()
    if loss_name == 'mae':
        return mae_loss
    elif loss_name == 'mse':
        return mse_loss
    elif loss_name == 'rmse':
        return rmse_loss
    elif loss_name == 'smape':
        return smape_loss
    elif loss_name == 'mape':
        return mape_loss
    elif loss_name == 'wmape':
        return wmape_loss
    else:
        raise ValueError(f"Unsupported loss function: {loss_name}")
